// Source file for the scene viewer program



// RING include files

#include "R3Graphics/R3Graphics.h"



// OpenGL include files 

#include <GL/glut.h>



// GLUT variables 

static int GLUTwindow = 0;
static int GLUTwindow_height = 480;
static int GLUTwindow_width = 640;
static int GLUTmouse[2] = { 0, 0 };
static int GLUTbutton[3] = { 0, 0, 0 };
static int GLUTarrows[4] = { 0, 0, 0, 0 };
static int GLUTmodifiers = 0;



// Display variables

static R3Viewer *viewers[2] = { NULL, NULL };
static int view = 1;
static float scale = 1.0;
static int first_mouse = 1;



// Scene variables

static char *scenename = NULL;
static R3Scene *scene = NULL;



// Interactive display options

static RNBoolean show_faces = TRUE;
static RNBoolean show_edges = FALSE;
static RNBoolean show_rays = FALSE;



void GLUTRedraw(void)
{
  // Set viewing transformation
  viewers[view]->Camera().Load();

  // Clear window 
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // Set lights
  static GLfloat light0_position[] = { 3.0, 4.0, 5.0, 0.0 };
  glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
  static GLfloat light1_position[] = { -3.0, -2.0, -3.0, 0.0 };
  glLightfv(GL_LIGHT1, GL_POSITION, light1_position);

  // Draw scene
  if (show_faces) {
    glEnable(GL_LIGHTING);
    scene->Draw();
  }
    
  // Draw wireframe
  if (show_edges) {
    glDisable(GL_LIGHTING);
    glColor3f(1.0, 0.0, 0.0);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    scene->Draw();
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  }

  // Draw rays from viewer[0]
  if (show_rays) {
    // Get some convenient variables
    const R3Camera *camera = &(viewers[0]->Camera());
    const R2Viewport *viewport = &(viewers[0]->Viewport());

    // Show camera
    glDisable(GL_LIGHTING);
    RNLoadRgb(0.0, 1.0, 0.0);
    camera->Outline();

    // Show some ray intersections
    int istep = viewport->Width()/8;
    int jstep = viewport->Height()/8;
    for (int i = istep/2; i < viewport->Width(); i += istep) {
      for (int j = jstep/2; j < viewport->Height(); j += jstep) {
        // Generate ray
        RNScalar dx = (2.0 * i - viewport->Width()) / viewport->Width();
        RNScalar dy = (2.0 * j - viewport->Height()) / viewport->Height();
        R3Point far_org = camera->Origin() + camera->Towards() * camera->Far();
        R3Vector far_right = camera->Right() * camera->Far() * tan(camera->XFOV());
        R3Vector far_up = camera->Up() * camera->Far() * tan(camera->YFOV());
        R3Point far_point = far_org + (far_right * dx) + (far_up * dy);
        R3Ray ray(camera->Origin(), far_point);
    
        // Find intersection
        R3Node *node;
        R3Point point;
        R3Vector normal;
        RNScalar t;
        if (scene->Intersects(ray, &node, &point, &normal, &t)) {
          // Draw sphere at intersection point
          glEnable(GL_LIGHTING);
          R3yellow_brdf.Draw();
          R3Sphere(point, 0.5*scale).Draw();

          // Draw normal at intersection point
          glDisable(GL_LIGHTING);
          RNLoadRgb(0.0, 1.0, 0.0);
          R3Span(point, point + normal * scale).Draw();

          // Draw span between eye and intersection point
          RNLoadRgb(1.0, 1.0, 0.0);
          R3Span(camera->Origin(), point).Draw();
        }
      }
    }

    // Reset OpenGL modes
    glEnable(GL_LIGHTING);
  }

  // Swap buffers 
  glutSwapBuffers();
}



static void
GLUTIdle(void)
{
    // Get mouse position 
    int x = GLUTmouse[0];
    int y = GLUTmouse[1];

    // Get mouse movement
    int dx, dy;
    static int mouse[2] = { 320, 240 };
    if (first_mouse) {
      dx = 0;
      dy = 0;
      first_mouse = 0;
    }
    else {
      dx = x - mouse[0];
      dy = y - mouse[1];
    }

    // Navigate the camera
    RNBoolean redraw = FALSE;
    if ((view == 0) || (GLUTmodifiers & GLUT_ACTIVE_SHIFT)) {
      // Initialize navigation parameters
      RNScalar pitch_factor = 1.0;
      RNScalar yaw_factor = pitch_factor;
      RNScalar step_factor = scale * pitch_factor;

      // Camera in hand navigation
      int relax_yaw = !(GLUTmodifiers & GLUT_ACTIVE_CTRL);
      if (!(GLUTmodifiers & GLUT_ACTIVE_CTRL)) yaw_factor = 0.0;
      if (GLUTbutton[1]) step_factor = 0.0;
      else if (GLUTbutton[2]) step_factor = -step_factor;
      viewers[0]->FlyCamera(pitch_factor, yaw_factor, 0.0, 
                            step_factor, relax_yaw, FALSE, x, y, dx, dy);
      redraw = TRUE;
    }
    else {
      // World in hand navigation 
      R3Point origin = scene->BBox().Centroid();
      if (GLUTbutton[0]) viewers[view]->RotateWorld(1.0, origin, x, y, dx, dy);
      else if (GLUTbutton[1]) viewers[view]->ScaleWorld(1.0, origin, x, y, dx, dy);
      else if (GLUTbutton[2]) viewers[view]->TranslateWorld(1.0, origin, x, y, dx, dy);
      redraw = TRUE;
    }

    /* Redraw */
    if (redraw) GLUTRedraw();

    /* Remember mouse coordinates */
    mouse[0] = x;
    mouse[1] = y;
}



void GLUTStop(void)
{
  // Destroy window 
  glutDestroyWindow(GLUTwindow);

  // Exit
  exit(0);
}



void GLUTResize(int w, int h)
{
  // Resize window
  glViewport(0, 0, w, h);

  /* Reset viewers */
  viewers[0]->ResizeViewport(0, 0, w, h);
  viewers[1]->ResizeViewport(0, 0, w, h);

  // Remember window size 
  GLUTwindow_width = w;
  GLUTwindow_height = h;

  // Redraw
  GLUTRedraw();
}



void GLUTMotion(int x, int y)
{
  // Invert y coordinate
  y = GLUTwindow_height - y;
  
  // Remember mouse position 
  GLUTmouse[0] = x;
  GLUTmouse[1] = y;
}



void GLUTMouse(int button, int state, int x, int y)
{
  // Invert y coordinate
  y = GLUTwindow_height - y;
  
  // Remember button state 
  int b = (button == GLUT_LEFT_BUTTON) ? 0 : ((button == GLUT_MIDDLE_BUTTON) ? 1 : 2);
  GLUTbutton[b] = (state == GLUT_DOWN) ? 1 : 0;

  // Remember modifiers 
  GLUTmodifiers = glutGetModifiers();

  // Remember mouse position 
  GLUTmouse[0] = x;
  GLUTmouse[1] = y;

  // Idle if any button is down
  if (GLUTbutton[0] || GLUTbutton[1] || GLUTbutton[2]) glutIdleFunc(GLUTIdle);
  else glutIdleFunc(0);

  // Set flags telling that mouse button was pressed
  first_mouse = 1;
}



void GLUTSpecial(int key, int x, int y)
{
  // Invert y coordinate
  y = GLUTwindow_height - y;

  // Process keyboard button event 

  // Remember mouse position 
  GLUTmouse[0] = x;
  GLUTmouse[1] = y;

  // Remember modifiers 
  GLUTmodifiers = glutGetModifiers();

  // Redraw
  GLUTRedraw();
}



void GLUTKeyboard(unsigned char key, int x, int y)
{
  // Flag to determine whether to redraw
  int redraw = 0;

  // Process keyboard button event 
  switch (key) {
  case 'C': 
  case 'c': {
    // Print camera
    const R3Camera& camera = viewers[view]->Camera();
    printf("#camera  %g %g %g  %g %g %g  %g %g %g  %g \n",
           camera.Origin().X(), camera.Origin().Y(), camera.Origin().Z(),
           camera.Towards().X(), camera.Towards().Y(), camera.Towards().Z(),
           camera.Up().X(), camera.Up().Y(), camera.Up().Z(),
           camera.YFOV());
    }
    break;
      
  case 'E':
  case 'e':
    show_edges = !show_edges;
    break;

  case 'F':
  case 'f':
    show_faces = !show_faces;
    break;

  case 'R':
  case 'r':
    show_rays = !show_rays;
    break;

  case 'V':
  case 'v':
    view = 1 - view;
    break;

  case 27: // ESCAPE
    GLUTStop();
    break;
  }

  // Redraw 
  if (redraw) GLUTRedraw();

  // Remember mouse position 
  GLUTmouse[0] = x;
  GLUTmouse[1] = GLUTwindow_height - y;

  // Remember modifiers 
  GLUTmodifiers = glutGetModifiers();

  // Redraw
  GLUTRedraw();
}



void GLUTInit(int *argc, char **argv)
{
  // Open window 
  glutInit(argc, argv);
  glutInitWindowPosition(100, 100);
  glutInitWindowSize(GLUTwindow_width, GLUTwindow_height);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH); // | GLUT_STENCIL
  GLUTwindow = glutCreateWindow("OpenGL Viewer");

  // Initialize GLUT callback functions 
  glutReshapeFunc(GLUTResize);
  glutDisplayFunc(GLUTRedraw);
  glutKeyboardFunc(GLUTKeyboard);
  glutSpecialFunc(GLUTSpecial);
  glutMouseFunc(GLUTMouse);
  glutMotionFunc(GLUTMotion);
  glutIdleFunc(0);
    
  // Initialize background color 
  glClearColor(0.6, 0.6, 0.6, 1.0);

  // Initialize lights */
  static GLfloat lmodel_ambient[] = { 0.2, 0.2, 0.2, 1.0 };
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
  static GLfloat light0_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
  glEnable(GL_LIGHT0);  static GLfloat light1_diffuse[] = { 0.5, 0.5, 0.5, 1.0 };
  glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);
  glEnable(GL_LIGHT1);
  glEnable(GL_NORMALIZE);
  glEnable(GL_LIGHTING);

  // Initialize graphics modes 
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
}



void GLUTMainLoop(void)
{
  // Run main loop -- never returns 
  glutMainLoop();
}



static int 
ParseArgs(int argc, char **argv)
{
    // Check number of arguments
    if ((argc == 2) && (*argv[1] == '-')) {
	printf("Usage: rdraw filename\n");
	exit(0);
    }

    // Parse arguments
    argc--; argv++;
    while (argc > 0) {
	if ((*argv)[0] == '-') {
  	    if (!strcmp(*argv, "-ray")) { argv++; argc--; scenename = *argv; }
	    else RNAbort("Invalid program argument: %s", *argv);
	    argv++; argc--;
	}
	else {
	    if (!scenename) scenename = *argv;
	    else RNAbort("Invalid program argument: %s", *argv);
	    argv++; argc--;
        }
    }

    // Check arguments
    if (!scenename) {
	RNFail("You did not specify a file name");	
	return FALSE;
    }

    // Return OK status 
    return TRUE;
}



static R3Viewer *
CreateInteriorViewer(R3Scene *scene)
{
    // Use camera in scene
    R2Viewport viewport(0, 0, GLUTwindow_width, GLUTwindow_height);
    return new R3Viewer(*(scene->Camera()), viewport);
}



static R3Viewer *
CreateBirdsEyeViewer(R3Scene *scene)
{
    // Setup camera view looking down the Z axis
    R3Box bbox = scene->BBox();
    assert(!bbox.IsEmpty());
    RNLength r = bbox.DiagonalRadius();
    assert((r > 0.0) && RNIsFinite(r));
    scale = r / 40.0;
    R3Point origin = bbox.Centroid() + R3posz_vector * (8.0 * r);
    R3Camera camera(origin, R3negz_vector, R3posy_vector, 0.4, 0.4, 0.1 * r, 1000.0 * r);
    R2Viewport viewport(0, 0, GLUTwindow_width, GLUTwindow_height);
    return new R3Viewer(camera, viewport);
}



int 
main(int argc, char **argv)
{
  // Initialize RING
  if (!R3InitGraphics()) exit(1);

  // Initialize GLUT
  GLUTInit(&argc, argv);

  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(1);

  // Read scene
  scene = R3ReadRayFile(scenename);
  if (!scene) exit(1);

  // Create viewers
  viewers[0] = CreateInteriorViewer(scene);
  viewers[1] = CreateBirdsEyeViewer(scene);

  // Run GLUT interface
  GLUTMainLoop();

  // Stop GLUT interface
  GLUTStop();

  // Stop RING
  R3StopGraphics();

  // Return success 
  return 0;
}











